/**
 * Custom logging to create more useful output.
 * Simply include this file at the top of other files to get extra logging logic.
 *
 * Log Levels:
 *  * 10 - DEBUG
 *  * 20 - INFO
 *  * 30 - WARNING
 *  * 40 - ERROR
 *
 * https://git.brandon-rodriguez.com/javascript/custom_logger
 * Version 1.1
 */


// System Imports.
const chalk = require('chalk');
const fs = require('fs');
const path = require('path');
const { Console } = require('console');

// User Imports.
const { getCurrentDate } = require('./helper_functions.js');


/**
 * "Fake" logging class used exclusively to initialize our real logging class.
 * Helps keep our real singleton class private.
 *
 * This is necessary because JavaScript doesn't seem to allow constructor overloading or private constructors.
 */
class Logging {

    /**
     * Class Constructor.
     */
    constructor() {
        throw new Error('Initialize class with "Logging.init()".');
    }

    /**
     * Method to initialize our singleton class.
     */
    static init(logging_dir) {
        if ( ! Logging.instance ) {
            Logging.instance = new LoggingClass(logging_dir);
        }
        return Logging.instance;
    }
}


/**
 * Real logging class.
 *
 * To extend class with additional logging levels:
 *  * Define a new "#x_file_logger" at the top of the file, to hold logging file descriptors.
 *  * In the constructor, copy one of the existing "this.#x_file_logger" sections and adjust for the new log level.
 *  * Copy one of the existing log level methods and adjust it for the new log level.
 *  * Finally, if applicable, add the new LogLeveLNum handling to the "#logToFile" method.
 */
class LoggingClass {

    // Define what will ultimately be our log file handlers.
    #debug_file_logger;
    #info_file_logger;
    #warning_file_logger;
    #error_file_logger;

    /**
     * Class Constructor.
     */
    constructor(logging_dir) {

        // First validate passed logging value. Start by checking if string.
        if (typeof logging_dir === 'string' || logging_dir instanceof String) {
            // Is string. Validate.
            logging_dir = logging_dir.trim();

            // Check last character.
            if (logging_dir.slice(-1) != '/') {
                logging_dir += '/';
            }

        } else {
            // Is not string. Raise error.
            throw new Error('Logging class must be initialized with directory location.');
        }

        // Attempt to create logging folder.
        fs.mkdir(logging_dir, err => {
            // Ignore "dir already exists" errors. Raise all others.
            if (err && err.code != 'EEXIST') {
                throw err;
            }
        });

        // Initialize DEBUG file logging stream.
        this.#debug_file_logger = new Console({
            stdout: fs.createWriteStream(logging_dir.concat('debug.log'), {'flags': 'a'}),
            stderr: fs.createWriteStream(logging_dir.concat('error.log'), {'flags': 'a'}),
            ignoreErrors: true,
            colorMode: false
        });

        // Initialize INFO file logging stream.
        this.#info_file_logger = new Console({
            stdout: fs.createWriteStream(logging_dir.concat('info.log'), {'flags': 'a'}),
            stderr: fs.createWriteStream(logging_dir.concat('error.log'), {'flags': 'a'}),
            ignoreErrors: true,
            colorMode: false
        });

        // Initialize WARNING file logging stream.
        this.#warning_file_logger = new Console({
            stdout: fs.createWriteStream(logging_dir.concat('warning.log'), {'flags': 'a'}),
            stderr: fs.createWriteStream(logging_dir.concat('error.log'), {'flags': 'a'}),
            ignoreErrors: true,
            colorMode: false
        });

        // Initialize ERROR file logging stream.
        this.#error_file_logger = new Console({
            stdout: fs.createWriteStream(logging_dir.concat('error.log'), {'flags': 'a'}),
            ignoreErrors: true,
            colorMode: false
        });
    }

    /**
     * Returns calling file and line num location.
     */
    getCallerInfo() {
        let originalPrepareStackTrace = Error.prepareStackTrace;
        Error.prepareStackTrace = (_, stack) => stack;
        let callee = new Error().stack[2];
        Error.prepareStackTrace = originalPrepareStackTrace;
        let relativeFileName = path.relative(process.cwd(), callee.getFileName());
        return [relativeFileName, callee.getLineNumber()]
    }

    /**
     * LOG logging method.
     */
    log(firstArgument, ...otherArguments) {
        // LOG level does not get prepend text. Nor does it save to file.
        // This is basically intended for "console only" output.
        console.log(firstArgument, ...otherArguments);
    }

    /**
     * DEBUG logging method.
     */
    debug(firstArgument, ...otherArguments) {
        // Get calling file name and line number.
        let [relativeFileName, lineNum] = this.getCallerInfo();

        // // Generate and print Console text.
        let prefix = ` ${chalk.green('[DEBUG]')} [${relativeFileName} ${lineNum}] `;
        let console_string = prefix.concat(firstArgument, ...otherArguments);
        console.log(console_string);

        // Generate and print Log File text.
        prefix = `${getCurrentDate()} [DEBUG] [${relativeFileName} ${lineNum}] `;
        let log_string = prefix.concat(firstArgument, ...otherArguments);
        this.#logToFile(log_string, 10);
    }

    /**
     * INFO logging method.
     */
    info(firstArgument, ...otherArguments) {
        // Get calling file name and line number.
        let [relativeFileName, lineNum] = this.getCallerInfo();

        // // Generate and print Console text.
        let prefix = ` ${chalk.blue('[INFO]')}  [${relativeFileName} ${lineNum}] `;
        let console_string = prefix.concat(firstArgument, ...otherArguments);
        console.log(console_string);

        // Generate and print Log File text.
        prefix = `${getCurrentDate()} [INFO] [${relativeFileName} ${lineNum}] `;
        let log_string = prefix.concat(firstArgument, ...otherArguments);
        this.#logToFile(log_string, 20);
    }

    /**
     * Alias function for WARNING level.
     */
    warn(firstArgument, ...otherArguments) {
        // Get calling file name and line number.
        let [relativeFileName, lineNum] = this.getCallerInfo();

        // // Generate and print Console text.
        let prefix = ` ${chalk.yellow('[WARN]')}  [${relativeFileName} ${lineNum}] `;
        let console_string = prefix.concat(firstArgument, ...otherArguments);
        console.log(console_string);

        // Generate and print Log File text.
        prefix = `${getCurrentDate()} [WARN] [${relativeFileName} ${lineNum}] `;
        let log_string = prefix.concat(firstArgument, ...otherArguments);
        this.#logToFile(log_string, 30);
    }

    /**
     * WARNING logging method.
     */
    warning(firstArgument, ...otherArguments) {
        // Get calling file name and line number.
        let [relativeFileName, lineNum] = this.getCallerInfo();

        // // Generate and print Console text.
        let prefix = ` ${chalk.yellow('[WARN]')}  [${relativeFileName} ${lineNum}] `;
        let console_string = prefix.concat(firstArgument, ...otherArguments);
        console.log(console_string);

        // Generate and print Log File text.
        prefix = `${getCurrentDate()} [WARN] [${relativeFileName} ${lineNum}] `;
        let log_string = prefix.concat(firstArgument, ...otherArguments);
        this.#logToFile(log_string, 30);
    }

    /**
     * ERROR logging method.
     */
    error(firstArgument, ...otherArguments) {
        // Get calling file name and line number.
        let [relativeFileName, lineNum] = this.getCallerInfo();

        // // Generate and print Console text.
        let prefix = ` ${chalk.red('[ERROR]')} [${relativeFileName} ${lineNum}] `;
        let console_string = prefix.concat(firstArgument, ...otherArguments);
        console.log(console_string);

        // Generate and print Log File text.
        prefix = `${getCurrentDate()} [ERROR] [${relativeFileName} ${lineNum}] `;
        let log_string = prefix.concat(firstArgument, ...otherArguments);
        this.#logToFile(log_string, 40);
    }

    /**
     * Writes provided string to log files.
     */
    #logToFile(log_string, log_level) {
        let directory = './logging/';
        let file_location = './logging/info.log';

        // Attempt to create logging folder.
        fs.mkdir(directory, err => {
            // Ignore "dir already exists" errors. Raise all others.
            if (err && err.code != 'EEXIST') {
                throw err;
            }
        });

        // Attempt to write to appropriate files, based on logging level.
        if (log_level >= 40) {
            // Error level logging or higher.
            this.#error_file_logger.log(log_string);
        }

        if (log_level >= 30) {
            // Warning level logging or higher.
            this.#warning_file_logger.log(log_string);
        }

        if (log_level >= 20) {
            // Info level logging or higher.
            this.#info_file_logger.log(log_string);
        }

        if (log_level >= 10) {
            // Debug level logging or higher.
            this.#debug_file_logger.log(log_string);
        }
    }

}


/**
 * Simple test to make sure all loging levels work as expected.
 */
function testLogLevels(logger) {
    logger.debug('DEBUG level test');
    logger.info('INFO level test.');
    logger.warn('WARN level test.');
    logger.warning('WARNING level test.');
    logger.error('ERROR level test.');
    logger.log('Log/default level test.');
}


module.exports = {
    Logging,
    testLogLevels,
};
